// 
// Copyright (c) 2010 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Net.Pop3.Protocol;
using Smdn.Net.Pop3.Protocol.DataStructures;
using Smdn.Net.Pop3.Protocol.Transmission;

namespace Smdn.Net.Pop3.Client {
  public sealed partial class PopMessage {
    public PopClient Client {
      get { ThrowIfSessionClosed(); return client; }
      internal set { client = value; }
    }

    public long MessageNumber {
      get { ThrowIfSessionClosed(); return scanList.MessageNumber; }
    }

    public long Length {
      get { ThrowIfSessionClosed(); return scanList.SizeInOctets; }
    }

    public string UniqueId {
      get
      {
        if (TryGetUniqueId(out var uniqueId))
          return uniqueId;
        else
          return null;
      }
    }

    internal PopScanListing ScanList {
      get { return scanList; }
      set { scanList = value; }
    }

    internal PopUniqueIdListing? UniqueIdList {
      get { return uniqueIdList; }
      set
      {
        uniqueIdList = value;
        isUniqueIdListed = true;
      }
    }

    internal string ListedUniqueId {
      get
      {
        if (uniqueIdList.HasValue && uniqueIdList.Value.UniqueId != null)
          return uniqueIdList.Value.UniqueId;
        else
          return null;
      }
    }

    public bool IsMarkedAsDeleted {
      get { ThrowIfSessionClosed(); return isMarkedAsDeleted; }
      internal set { isMarkedAsDeleted = value; }
    }

    internal PopMessage(PopClient client, PopScanListing scanList, PopUniqueIdListing? uniqueIdList)
    {
      this.client = client;
      this.Body = new MessageBody(this);
      this.scanList = scanList;
      this.uniqueIdList = uniqueIdList;
      this.isMarkedAsDeleted = false;
      this.isUniqueIdListed = uniqueIdList.HasValue;
    }

    private PopClient client;
    private PopScanListing scanList;
    private PopUniqueIdListing? uniqueIdList;
    private bool isMarkedAsDeleted;
    private bool isUniqueIdListed;

    /*
     * operation
     */
    public void MarkAsDeleted()
    {
      ThrowIfSessionClosed();

      client.RunExclusively(MarkAsDeletedCore);
    }

    private void MarkAsDeletedCore()
    {
      if (isMarkedAsDeleted)
        return;

      client.Delete(this);

      isMarkedAsDeleted = true;
    }

    public string GetUniqueId()
    {
      if (TryGetUniqueId(out var uniqueId))
        return uniqueId;
      else
        throw new PopIncapableException(PopCapability.Uidl);
    }

    public bool TryGetUniqueId(out string uniqueId)
    {
      ThrowIfSessionClosed();

      if (isUniqueIdListed) {
        uniqueId = ListedUniqueId;
      }
      else {
        ThrowIfMarkedAsDeleted();

        uniqueId = client.RunExclusively<string>(ListUniqueId);
      }

      return (uniqueId != null);
    }

    private string ListUniqueId()
    {
#if DEBUG
      client.ThrowIfNotInExclusiveOperation();
#endif

      var ret = client.Session.Uidl(MessageNumber, out var uid);

      if (ret.IsResponseStatus(PopStatusIndicator.Positive)) {
        isUniqueIdListed = true;
        uniqueIdList = uid;
        return uid.UniqueId;
      }

      if (ret.IsResponseStatus(PopStatusIndicator.Negative)) {
        isUniqueIdListed = true;
        uniqueIdList = null;
        return null;
      }

      PopClient.ThrowIfError(ret);
      return null; // never return
    }

    internal const int EntireMessage = -1;

    /*
     * OpenRead()
     */
    /// <param name="maxLines">
    /// A maximum line count to retrieve.
    /// <list type="number">
    ///   <listheader><term>value</term><description>behavior</description></listheader>
    ///   <item><term>-1</term><description>entire message</description></item>
    ///   <item><term>0</term><description>message without body</description></item>
    ///   <item><term>1 or greater</term><description>message with specified lines of body</description></item>
    /// </list>
    /// </param>
    public Stream OpenRead(
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.OpenRead(maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    public Task<Stream> OpenReadAsync(
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.OpenReadAsync(maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    /*
     * ReadAs<TResult>(Converter<Stream, TResult>)
     */
    public TResult ReadAs<TResult>(
      Converter<Stream, TResult> converter,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.ReadAs(converter: converter, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    public Task<TResult> ReadAsAsync<TResult>(
      Converter<Stream, Task<TResult>> converter,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.ReadAsAsync(converter: converter, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    /*
     * ReadAs<TResult>(Func<Stream, TArg, TResult>)
     */
    public TResult ReadAs<TArg, TResult>(
      Func<Stream, TArg, TResult> converter,
      TArg arg,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.ReadAs(converter: converter, arg: arg, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    public Task<TResult> ReadAsAsync<TArg, TResult>(
      Func<Stream, TArg, Task<TResult>> converter,
      TArg arg,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.ReadAsAsync(converter: converter, arg: arg, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    /*
     * ReadAsByteArray()
     */
    public byte[] ReadAsByteArray(
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.ReadAsByteArray(maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    public Task<byte[]> ReadAsByteArrayAsync(
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.ReadAsByteArrayAsync(maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    /*
     * WriteTo(Stream)
     */
    public void WriteTo(
      Stream stream,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.WriteTo(stream: stream, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    public Task WriteToAsync(
      Stream stream,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.WriteToAsync(stream: stream, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    /*
     * WriteTo(BinaryWriter)
     */
    public void WriteTo(
      BinaryWriter writer,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.WriteTo(writer: writer, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    public Task WriteToAsync(
      BinaryWriter writer,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.WriteToAsync(writer: writer, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    /*
     * Save()
     */
    public void Save(
      string path,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.Save(path: path, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    public Task SaveAsync(
      string path,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.SaveAsync(path: path, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    /*
     * OpenText()
     */
    public StreamReader OpenText(
      Encoding charset = null,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.OpenText(charset: charset, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    public Task<StreamReader> OpenTextAsync(
      Encoding charset = null,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.OpenTextAsync(charset: charset, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    /*
     * ReadAs<TResult>(Converter<StreamReader, TResult>)
     */
    public TResult ReadAs<TResult>(
      Converter<StreamReader, TResult> converter,
      Encoding charset = null,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.ReadAs(converter: converter, charset: charset, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    public Task<TResult> ReadAsAsync<TResult>(
      Converter<StreamReader, Task<TResult>> converter,
      Encoding charset = null,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.ReadAsAsync(converter: converter, charset: charset, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    /*
     * ReadAs<TResult>(Func<StreamReader, TArg, TResult>)
     */
    public TResult ReadAs<TArg, TResult>(
      Func<StreamReader, TArg, TResult> converter,
      TArg arg,
      Encoding charset = null,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.ReadAs(converter: converter, arg: arg, charset: charset, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    public Task<TResult> ReadAsAsync<TArg, TResult>(
      Func<StreamReader, TArg, Task<TResult>> converter,
      TArg arg,
      Encoding charset = null,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.ReadAsAsync(converter: converter, arg: arg, charset: charset, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    /*
     * ReadAsString()
     */
    public string ReadAsString(
      Encoding charset = null,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.ReadAsString(charset: charset, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    public Task<string> ReadAsStringAsync(
      Encoding charset = null,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.ReadAsStringAsync(charset: charset, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    /*
     * ReadAsStringLines()
     */
    public IReadOnlyList<string> ReadAsStringLines(
      Encoding charset = null,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.ReadAsStringLines(charset: charset, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    public Task<IReadOnlyList<string>> ReadAsStringLinesAsync(
      Encoding charset = null,
      int maxLines = EntireMessage,
      CancellationToken cancellationToken = default,
      IProgress<PopMessageRetrieveProgressInfo> progress = null
    ) =>
      Body.ReadAsStringLinesAsync(charset: charset, maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    #region "obsolete read operations"
    /*
     * ReadAs<TResult>(Converter<Stream>)
     */
    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public TResult ReadAs<TResult>(Converter<Stream, TResult> converter)
      => Body.ReadAs(converter: converter);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public TResult ReadAs<TResult>(int maxLines, Converter<Stream, TResult> converter)
      => Body.ReadAs(maxLines: maxLines, converter: converter);

    /*
     * ReadAs<TResult>(Func<Stream, ...>)
     */
    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public TResult ReadAs<T, TResult>
      (Func<Stream, T, TResult> read, T arg)
      => Body.ReadAs(converter: read, arg: arg);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public TResult ReadAs<T1, T2, TResult>
      (Func<Stream, T1, T2, TResult> read, T1 arg1, T2 arg2)
      => Body.ReadAs(converter: (s, a) => read(s, a.arg1, a.arg2), arg: (arg1, arg2));

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public TResult ReadAs<T1, T2, T3, TResult>
      (Func<Stream, T1, T2, T3, TResult> read, T1 arg1, T2 arg2, T3 arg3)
      => Body.ReadAs(converter: (s, a) => read(s, a.arg1, a.arg2, a.arg3), arg: (arg1, arg2, arg3));

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public TResult ReadAs<T1, T2, T3, T4, TResult>
      (Func<Stream, T1, T2, T3, T4, TResult> read, T1 arg1, T2 arg2, T3 arg3, T4 arg4)
      => Body.ReadAs(converter: (s, a) => read(s, a.arg1, a.arg2, a.arg3, arg4), arg: (arg1, arg2, arg3, arg4));

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public TResult ReadAs<T, TResult>
      (int maxLines,
       Func<Stream, T, TResult> read, T arg)
      => Body.ReadAs(maxLines: maxLines, converter: read, arg: arg);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public TResult ReadAs<T1, T2, TResult>
      (int maxLines,
       Func<Stream, T1, T2, TResult> read, T1 arg1, T2 arg2)
      => Body.ReadAs(maxLines: maxLines, converter: (s, a) => read(s, a.arg1, a.arg2), arg: (arg1, arg2));

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public TResult ReadAs<T1, T2, T3, TResult>
      (int maxLines,
       Func<Stream, T1, T2, T3, TResult> read, T1 arg1, T2 arg2, T3 arg3)
      => Body.ReadAs(maxLines: maxLines, converter: (s, a) => read(s, a.arg1, a.arg2, a.arg3), arg: (arg1, arg2, arg3));

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public TResult ReadAs<T1, T2, T3, T4, TResult>
      (int maxLines,
       Func<Stream, T1, T2, T3, T4, TResult> read, T1 arg1, T2 arg2, T3 arg3, T4 arg4)
      => Body.ReadAs(maxLines: maxLines, converter: (s, a) => read(s, a.arg1, a.arg2, a.arg3, a.arg4), arg: (arg1, arg2, arg3, arg4));

    /*
     * ReadAllBytes()
     */
    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public byte[] ReadAllBytes()
      => Body.ReadAsByteArray();

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public byte[] ReadAllBytes(int maxLines)
      => Body.ReadAsByteArray(maxLines: maxLines);

    /*
     * WriteTo(Stream)
     */

    /*
     * WriteTo(BinaryWriter)
     */

    /*
     * Save()
     */

    /*
     * OpenText()
     */
    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public StreamReader OpenText(int maxLines)
      => Body.OpenText(maxLines: maxLines);

    /*
     * ReadAs<TResult>(Converter<StreamReader>)
     */
    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public TResult ReadAs<TResult>(Converter<StreamReader, TResult> converter)
      => Body.ReadAs(converter: converter);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public TResult ReadAs<TResult>(Encoding encoding, Converter<StreamReader, TResult> converter)
      => Body.ReadAs(converter: converter, charset: encoding);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public TResult ReadAs<TResult>(int maxLines, Converter<StreamReader, TResult> converter)
      => Body.ReadAs(converter: converter, maxLines: maxLines);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public TResult ReadAs<TResult>(Encoding encoding, int maxLines, Converter<StreamReader, TResult> converter)
      => Body.ReadAs(converter: converter, charset: encoding, maxLines: maxLines);

    /*
     * ReadLines()
     */
    [Obsolete("use ReadAsStringLines() instead")]
    public IEnumerable<string> ReadLines()
      => Body.ReadAsStringLines();

    [Obsolete("use ReadAsStringLines() instead")]
    public IEnumerable<string> ReadLines(Encoding encoding)
      => Body.ReadAsStringLines(charset: encoding);

    [Obsolete("use ReadAsStringLines() instead")]
    public IEnumerable<string> ReadLines(int maxLines)
      => Body.ReadAsStringLines(maxLines: maxLines);

    [Obsolete("use ReadAsStringLines() instead")]
    public IEnumerable<string> ReadLines(Encoding encoding, int maxLines)
      => Body.ReadAsStringLines(charset: encoding, maxLines: maxLines);

    /*
     * ReadAllLines()
     */
    [Obsolete("use ReadAsStringLines() instead")]
    public string[] ReadAllLines()
      => Body.ReadAsStringLines().ToArray();

    [Obsolete("use ReadAsStringLines() instead")]
    public string[] ReadAllLines(Encoding encoding)
      => Body.ReadAsStringLines(charset: encoding).ToArray();

    [Obsolete("use ReadAsStringLines() instead")]
    public string[] ReadAllLines(int maxLines)
      => Body.ReadAsStringLines(maxLines: maxLines).ToArray();

    [Obsolete("use ReadAsStringLines() instead")]
    public string[] ReadAllLines(Encoding encoding, int maxLines)
      => Body.ReadAsStringLines(charset: encoding, maxLines: maxLines).ToArray();

    /*
     * ReadAllText()
     */
    [Obsolete("use ReadAsString() instead")]
    public string ReadAllText()
      => Body.ReadAsString();

    [Obsolete("use ReadAsString() instead")]
    public string ReadAllText(Encoding encoding)
      => Body.ReadAsString(charset: encoding);

    [Obsolete("use ReadAsString() instead")]
    public string ReadAllText(int maxLines)
      => Body.ReadAsString(maxLines: maxLines);

    [Obsolete("use ReadAsString() instead")]
    public string ReadAllText(Encoding encoding, int maxLines)
      => Body.ReadAsString(charset: encoding, maxLines: maxLines);
    #endregion

    /*
     * utility methods
     */
    internal void ThrowIfMarkedAsDeleted()
    {
      if (IsMarkedAsDeleted)
        throw new PopMessageDeletedException(MessageNumber);
    }

    private void ThrowIfSessionClosed()
    {
      if (client == null)
        throw new InvalidOperationException("session has been closed");
    }

    private void ThrowIfMaxLinesOutOfRange(int maxLines)
    {
      if (maxLines < -1)
        throw new ArgumentOutOfRangeException(nameof(maxLines), maxLines, "must be -1, 0 or positive number");
    }

    public override string ToString()
    {
      return string.Format("{{PopMessage: Authority='{0}', MessageNumber={1}, Length={2}}}",
                           client.Profile.Authority,
                           MessageNumber,
                           Length);
    }
  }
}
