// 
// Copyright (c) 2010 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Net;
using System.Threading;

using Smdn.Net.Connection;
using Smdn.Net.Pop3.Protocol.Transmission;
using Smdn.Net.Pop3.Protocol.Transmission.Transaction;
using Smdn.Net.Pop3.Urls;
using Smdn.Security.Authentication.Sasl.Client;

namespace Smdn.Net.Pop3.Protocol.Session {
  partial class PopSession {
    /*
     * transaction methods : authorization state
     */
    private string lastUserCommandArgument = null; // used by USER and PASS command

    /// <summary>sends STLS command</summary>
    /// <remarks>valid in authorization state</remarks>
    public PopCommandResult Stls(UpgradeConnectionStreamCallback createAuthenticatedStream,
                                 bool reissueCapability = false)
    {
      RejectNonAuthorizationState();

      if (createAuthenticatedStream == null)
        throw new ArgumentNullException(nameof(createAuthenticatedStream));

      PopCommandResult result;

      using (var t = new StlsTransaction(connection, createAuthenticatedStream)) {
        if ((result = ProcessTransaction(t)).IsSuccess)
          // RFC 2595 Using TLS with IMAP, POP3 and ACAP
          // http://tools.ietf.org/html/rfc2595
          // 4. POP3 STARTTLS extension
          //              Once TLS has been started, the client MUST discard cached
          //              information about server capabilities and SHOULD re-issue
          //              the CAPA command.  This is necessary to protect against
          //              man-in-the-middle attacks which alter the capabilities list
          //              prior to STLS.  The server MAY advertise different
          //              capabilities after STLS.
          SetServerCapabilities(null);
        else
          return result;
      }

      if (reissueCapability)
        Capa();

      return result;
    }

    /// <summary>sends USER command</summary>
    /// <remarks>valid in authorization state</remarks>
    public PopCommandResult User(ICredentialsByHost credentials)
    {
      if (credentials == null)
        throw new ArgumentNullException(nameof(credentials));

      var credential = credentials.LookupCredential(connection, null, null);

      if (credential == null)
        return PopCommandResult.CreateFailure($"credential not found for {connection.Host}:{connection.Port}");

      return User(credential.UserName);
    }

    public PopCommandResult User(string username)
    {
      RejectNonConnectedState();

      if (State != PopSessionState.Authorization)
        return PopCommandResult.CreateSuccess("already authenticated");

      if (string.IsNullOrEmpty(username))
        return PopCommandResult.CreateFailure("username is empty");

      using (var t = new UserTransaction(connection)) {
        t.RequestArguments["name"] = username;

        if (ProcessTransaction(t).IsSuccess)
          lastUserCommandArgument = username;
        else
          lastUserCommandArgument = null;

        return t.Result;
      }
    }

    /// <summary>sends PASS command</summary>
    /// <remarks>valid in authorization state</remarks>
    public PopCommandResult Pass(ICredentialsByHost credentials)
    {
      if (credentials == null)
        throw new ArgumentNullException(nameof(credentials));

      return PassCore(credentials, null);
    }

    /// <summary>sends PASS command</summary>
    /// <remarks>valid in authorization state</remarks>
    public PopCommandResult Pass(string password)
    {
      return PassCore(null, password);
    }

    private PopCommandResult PassCore(ICredentialsByHost credentials, string password)
    {
      var ret = RejectNonConnectedOrGetAuthenticatedResult();

      if (ret != null)
        return ret;

      if (lastUserCommandArgument == null)
        throw new PopProtocolViolationException("issue USER command first");

      if (password == null) {
        var credential = credentials.LookupCredential(connection, lastUserCommandArgument, null);

        if (credential == null)
          return PopCommandResult.CreateFailure($"credential not found for {lastUserCommandArgument}@{connection.Host}:{connection.Port}");

        password = credential.Password;
      }

      if (string.IsNullOrEmpty(password))
        return PopCommandResult.CreateFailure("password is empty");

      using (var t = new PassTransaction(connection)) {
        t.RequestArguments["string"] = password;

        if (ProcessTransaction(t).IsSuccess) {
          UpdateAuthority(lastUserCommandArgument, null);
          TransitStateTo(PopSessionState.Transaction);
        }

        lastUserCommandArgument = null;

        return t.Result;
      }
    }

    /// <summary>sends USER and PASS command</summary>
    /// <remarks>valid in authorization state</remarks>
    public PopCommandResult Login(ICredentialsByHost credentials, string username = null)
    {
      RejectNonConnectedState();

      if (credentials == null)
        throw new ArgumentNullException(nameof(credentials));

      var credential = credentials.LookupCredential(connection, username, null);

      if (credential == null)
        return PopCommandResult.CreateFailure($"credential not found for {username}@{connection.Host}:{connection.Port}");

      return Login(credential.UserName, credential.Password);
    }

    /// <summary>sends USER and PASS command</summary>
    /// <remarks>valid in authorization state</remarks>
    public PopCommandResult Login(string username, string password)
    {
      var result = User(username);

      if (result.IsSuccess)
        return Pass(password);
      else
        return result;
    }

    /// <summary>sends APOP command</summary>
    /// <remarks>valid in authorization state</remarks>
    public PopCommandResult Apop(ICredentialsByHost credentials, string username = null)
    {
      if (credentials == null)
        throw new ArgumentNullException(nameof(credentials));

      var ret = RejectNonConnectedOrGetAuthenticatedResult();

      if (ret != null)
        return ret;

      if (!ApopAvailable)
        throw new PopIncapableException("server does not support APOP");

      ret = LookupAppropriateCredential(credentials, username, PopAuthenticationMechanism.Apop, out var credential);

      if (ret != null)
        return ret;
      if (string.IsNullOrEmpty(credential.UserName))
        return PopCommandResult.CreateFailure("username of credential is empty");
      if (string.IsNullOrEmpty(credential.Password))
        return PopCommandResult.CreateFailure("password of credential is empty");

      using (var t = new ApopTransaction(connection)) {
        t.RequestArguments["name"] = credential.UserName;
        t.RequestArguments["digest"] = ApopDigest.Calculate(timestamp, credential.Password);

        if (ProcessTransaction(t).IsSuccess) {
          UpdateAuthority(credential.UserName, PopAuthenticationMechanism.Apop);
          TransitStateTo(PopSessionState.Transaction);
        }

        return t.Result;
      }
    }

    /// <summary>sends AUTH command</summary>
    /// <remarks>valid in authorization state</remarks>
    public PopCommandResult Auth(PopAuthenticationMechanism authenticationMechanism,
                                 ICredentialsByHost credentials,
                                 string username = null,
                                 CancellationToken cancellationToken = default)
    {
      if (credentials == null)
        throw new ArgumentNullException(nameof(credentials));
      if (authenticationMechanism == null)
        throw new ArgumentNullException(nameof(authenticationMechanism));

      var ret = RejectNonConnectedOrGetAuthenticatedResult();

      if (ret != null)
        return ret;

      // TODO: check Request.Arguments, not here
      IsCapable(authenticationMechanism, throwIfIncapable: handlesIncapableAsException);

      ret = LookupAppropriateCredential(credentials, username, authenticationMechanism, out var credential);

      if (ret != null)
        return ret;

      using (var saslMechanism = SaslClientMechanism.Create(authenticationMechanism.Value)) {
        /*
         * http://tools.ietf.org/html/rfc5034
         * RFC 5034 - The Post Office Protocol (POP3) Simple Authentication and Security Layer (SASL) Authentication Mechanism
         * 4.  The AUTH Command
         *       The service name specified by this protocol's profile of SASL
         *       is "pop".
         */
        saslMechanism.ServiceName = "pop";
        saslMechanism.Credential = credential;

        using (var t = new AuthTransaction(connection,
                                           saslMechanism,
                                           cancellationToken)) {
          t.RequestArguments["mechanism"] = authenticationMechanism.Value;

          return AuthInternal(t,
                              saslMechanism.Credential?.UserName,
                              authenticationMechanism);
        }
      }
    }

    /// <summary>sends AUTH command</summary>
    /// <remarks>valid in authorization state</remarks>
    public PopCommandResult Auth(SaslClientMechanism specificAuthenticationMechanism,
                                 CancellationToken cancellationToken = default(CancellationToken))
    {
      if (specificAuthenticationMechanism == null)
        throw new ArgumentNullException(nameof(specificAuthenticationMechanism));

      var ret = RejectNonConnectedOrGetAuthenticatedResult();

      if (ret != null)
        return ret;

      using (var t = new AuthTransaction(connection,
                                         specificAuthenticationMechanism,
                                         cancellationToken)) {
        t.RequestArguments["mechanism"] = specificAuthenticationMechanism.Name;

        return AuthInternal(t,
                            specificAuthenticationMechanism.Credential?.UserName,
                            PopAuthenticationMechanism.GetKnownOrCreate(specificAuthenticationMechanism.Name));
      }
    }

    private PopCommandResult AuthInternal(AuthTransaction t,
                                          string username,
                                          PopAuthenticationMechanism authenticationMechanism)
    {
      PopCommandResult result;

      try {
        result = ProcessTransaction(t);
      }
      catch (AuthTransaction.SaslInitialExchangeFailedException ex) {
        return PopCommandResult.CreateFailure(ex.Message);
      }

      if (result.IsSuccess) {
        UpdateAuthority(username, authenticationMechanism);
        TransitStateTo(PopSessionState.Transaction);
      }

      return result;
    }

    private PopCommandResult RejectNonConnectedOrGetAuthenticatedResult()
    {
      RejectNonConnectedState();

      if (State == PopSessionState.Authorization)
        return null;
      else
        return PopCommandResult.CreateSuccess("already authenticated");
    }

    private PopCommandResult LookupAppropriateCredential(ICredentialsByHost credentials,
                                                         string username,
                                                         PopAuthenticationMechanism authenticationMechanism,
                                                         out NetworkCredential credential)
    {
      credential = credentials.LookupCredential(connection, username, authenticationMechanism?.Value);

      if (credential == null)
        return PopCommandResult.CreateFailure($"credential not found for {username};AUTH={authenticationMechanism.Value}@{connection.Host}:{connection.Port}");
      else
        return null;
    }

    internal void UpdateAuthority(string username, PopAuthenticationMechanism authType)
    {
      authority.Scheme    = connection.IsSecurePortConnection ? PopUri.UriSchemePops : PopUri.UriSchemePop;
      authority.Host      = connection.Host;
      authority.Port      = connection.Port;
      authority.UserName  = username;
      authority.AuthType  = authType;

      TraceInfo(string.Concat("authority: ", authority));
    }
  }
}
