// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Net.Connection;
using Smdn.Net.MessageAccessProtocols.Strings;
using Smdn.Net.Pop3.Protocol.Transmission;
using Smdn.Net.Pop3.Protocol.Transmission.Transaction;
using Smdn.Net.Pop3.Urls;

using Trace = Smdn.Net.Pop3.Client.Trace;

namespace Smdn.Net.Pop3.Protocol.Session {
  public sealed partial class PopSession : IConnectionInfo, IDisposable {
    string IConnectionInfo.Host { get { CheckDisposed(); return connection.Host; } }
    int IConnectionInfo.Port { get { CheckDisposed(); return connection.Port; } }

    /*
     * properties
     */
    internal int? Id {
      get { return (connection == null) ? (int?)null : connection.Id; }
    }

    public bool HandlesIncapableAsException {
      get { return handlesIncapableAsException; }
      set { handlesIncapableAsException = value; }
    }

    public int SendTimeout {
      get { CheckDisposed(); return connection.SendTimeout; }
      set { CheckDisposed(); connection.SendTimeout = value; }
    }

    public int ReceiveTimeout {
      get { CheckDisposed(); return connection.ReceiveTimeout; }
      set { CheckDisposed(); connection.ReceiveTimeout = value; }
    }

    public bool IsSecureConnection {
      get { CheckDisposed(); return connection.IsSecurePortConnection || connection.IsSecureConnection; }
    }

    public int TransactionTimeout {
      get { CheckDisposed(); return transactionTimeout; }
      set
      {
        CheckDisposed();
        if (value < -1)
          throw ExceptionUtils.CreateArgumentMustBeGreaterThanOrEqualTo(-1, nameof(TransactionTimeout), value);
        transactionTimeout = value;
      }
    }

    public bool IsTransactionProceeding {
      get
      {
        if (Monitor.TryEnter(transactionLockObject, 0)) {
          Monitor.Exit(transactionLockObject);
          return false;
        }
        else {
          return true;
        }
      }
    }

    /*
     * construction / destruction
     */
    public PopSession()
    {
    }

    public PopSession(string hostname,
                      int port = -1,
                      int transactionTimeout = DefaultTransactionTimeout,
                      int sendTimeout = DefaultSendTimeout,
                      int receiveTimeout = DefaultReceiveTimeout,
                      UpgradeConnectionStreamCallback createAuthenticatedStream = null)
    {
      if (transactionTimeout < -1)
        throw ExceptionUtils.CreateArgumentMustBeGreaterThanOrEqualTo(-1, nameof(transactionTimeout), transactionTimeout);

      this.transactionTimeout = transactionTimeout;
      this.SendTimeout = sendTimeout;
      this.ReceiveTimeout = receiveTimeout;

      using (var cts = new CancellationTokenSource(transactionTimeout)) {
        ConnectAsync(
          hostname: hostname,
          port: port,
          createAuthenticatedStream: createAuthenticatedStream,
          cancellationToken: cts.Token
        ).GetAwaiter().GetResult();
      }
    }

    public override string ToString()
    {
      CheckDisposed();

      return string.Format("{{Connection={0}, State={1}}}",
                           connection,
                           State);
    }

    private const int DefaultTransactionTimeout = Timeout.Infinite;
    private const int DefaultSendTimeout = Timeout.Infinite;
    private const int DefaultReceiveTimeout = Timeout.Infinite;

    private int transactionTimeout = DefaultTransactionTimeout;
    private bool handlesIncapableAsException = false;
    private readonly object transactionLockObject = new object();

    /*
     * session state properties
     */
    public PopSessionState State { get; private set; } = PopSessionState.NotConnected;

    public Uri Authority {
      get { CheckDisposed(); return authority == null ? null : authority.Uri; }
    }

    public IReadOnlyStringSet<PopCapability> ServerCapabilities {
      get { CheckDisposed(); return serverCapabilities; }
    }

    /// <summary>timestamp in server banner greeting</summary>
    public string Timestamp {
      get { CheckDisposed(); return timestamp; }
    }

    public bool ApopAvailable {
      get { return !string.IsNullOrEmpty(Timestamp); }
    }

    private PopConnection connection = new PopConnection();
    private PopUriBuilder authority = new PopUriBuilder();
    private IReadOnlyStringSet<PopCapability> serverCapabilities = PopCapabilitySet.ReadOnlyEmpty;
    private string timestamp = null;

    /*
     * transaction methods : connect/disconnect
     */
    public Task ConnectAsync(string hostname,
                             int port = -1,
                             ResolveHostAddressCallback resolveHostAddress = null,
                             UpgradeConnectionStreamCallback createAuthenticatedStream = null,
                             CancellationToken cancellationToken = default)
    {
      CheckDisposed();

      if (connection.IsConnected)
        throw new InvalidOperationException("already connected");

      return ConnectAsyncImpl();

      async Task ConnectAsyncImpl()
      {
        TraceInfo("connecting");

        await connection.ConnectAsync(
          hostname: hostname,
          port: port,
          resolveHostAddress: resolveHostAddress,
          createAuthenticatedStream: createAuthenticatedStream,
          cancellationToken: cancellationToken
        ).ConfigureAwait(false);

        TraceInfo("connected");

        using (var t = new GreetingTransaction(connection)) {
          try {
            if (!ProcessTransaction(t).IsSuccess) {
              CloseConnection();
              throw new PopConnectionException(string.Concat("session establishment failed: ", t.Result.Description));
            }
          }
          catch (PopMalformedResponseException ex) {
            throw new PopConnectionException("session establishment failed (malformed or invalid response)", ex);
          }
          catch (TimeoutException ex) {
            throw new PopConnectionException("connection timed out", ex);
          }

          timestamp = t.TransactionResult;

          TransitStateTo(PopSessionState.Authorization);
        }

        if (ApopAvailable)
          TraceInfo(string.Concat("APOP is available, timestamp is ", timestamp));
      }
    }

    /// <summary>disconnects session</summary>
    /// <remarks>this method sends CLOSE and LOGOUT before disconnect the session.</remarks>
    public void Disconnect(bool logout = true)
    {
      CheckDisposed();

      if (logout) {
        if (State == PopSessionState.Transaction || State == PopSessionState.Authorization)
          Quit();
      }

      CloseConnection();
    }

    public void Dispose() => CloseConnection();

    private void CloseConnection()
    {
      if (connection == null)
        return;

      try {
        TraceInfo("disconnecting");

        connection.Close();

        TraceInfo("disconnected");
      }
      finally {
        TransitStateTo(PopSessionState.Disconnected);

        connection = null;
      }
    }

    /*
     * transaction methods : generic use, exntension, etc.
     */

    /// <summary>sends generic/undefined command</summary>
    public PopCommandResult GenericCommand(string command, params string[] arguments)
    {
      return GenericCommand(command, false, arguments);
    }

    /// <summary>sends generic/undefined command</summary>
    public PopCommandResult GenericCommand(string command, bool isResponseMultiline, params string[] arguments)
    {
      return GenericCommand(command, isResponseMultiline, out _, arguments);
    }

    /// <summary>sends generic/undefined command</summary>
    public PopCommandResult GenericCommand(string command, out IReadOnlyList<PopResponse> responses, params string[] arguments)
    {
      return GenericCommand(command, false, out responses, arguments);
    }

    /// <summary>sends generic/undefined command</summary>
    public PopCommandResult GenericCommand(string command, bool isResponseMultiline, out IReadOnlyList<PopResponse> responses, params string[] arguments)
    {
      RejectNonConnectedState();

      responses = null;

      using (var t = new GenericCommandTransaction(connection, command, isResponseMultiline)) {
        if (arguments != null && 0 < arguments.Length)
          t.RequestArguments["arguments"] = string.Join(" ", arguments);

        if (ProcessTransaction(t).IsSuccess)
          responses = t.Result.ReceivedResponses;

        return t.Result;
      }
    }

    /*
     * methods for internal state management
     */
    internal void SetServerCapabilities(IReadOnlyStringSet<PopCapability> newCapabilities)
    {
      serverCapabilities = newCapabilities ?? PopCapabilitySet.ReadOnlyEmpty;
    }

    private void TransitStateTo(PopSessionState newState)
    {
      if (newState == State)
        return;

      switch (newState) {
        case PopSessionState.NotConnected:
          throw new InvalidOperationException("it is not allowed to transit to non-connected state");

        case PopSessionState.Authorization:
          UpdateAuthority(null, null);
          TraceInfo("now in authorization state");
          break;

        case PopSessionState.Transaction:
          TraceInfo("now in transaction state");
          if (State == PopSessionState.Authorization)
            TraceInfo("logged in");
          break;

        case PopSessionState.Update:
          TraceInfo("now in update state");
          break;

        case PopSessionState.Disconnected:
          authority = null;
          if (State == PopSessionState.Update || State == PopSessionState.Transaction)
            TraceInfo("logged out");
          TraceInfo("disconnected");
          break;
      }

      State = newState;
    }

    /*
     * methods for checking and validating session status
     */
    private void CheckDisposed()
    {
      if (State == PopSessionState.Disconnected)
        throw new ObjectDisposedException(GetType().FullName);
    }

    private void RejectNonConnectedState()
    {
      CheckDisposed();

      if (State == PopSessionState.NotConnected)
        throw new InvalidOperationException("session is not connected yet");
    }

    private void RejectNonAuthorizationState()
    {
      RejectNonConnectedState();

      if (State != PopSessionState.Authorization)
        throw new PopProtocolViolationException("session is not in authorization state");
    }

    private void RejectNonTransactionState()
    {
      RejectNonConnectedState();

      if (State != PopSessionState.Transaction)
        throw new PopProtocolViolationException("session is not in transaction state");
    }

    internal bool IsCapable(IPopExtension extension, bool throwIfIncapable)
    {
      var isCapable = serverCapabilities.IsCapable(extension);

      if (throwIfIncapable && !isCapable)
        throw new PopIncapableException(extension.RequiredCapability);

      return isCapable;
    }

    /*
    private void CheckServerCapability(PopCapability capability)
    {
      if (capability != null && !serverCapabilities.IsCapable(capability))
        throw new PopIncapableException(capability);
    }
    */

    /*
     * tracing
     */
    [System.Diagnostics.Conditional("TRACE")]
    private void TraceInfo(string message)
    {
      Trace.Log(this, message);
    }

    [System.Diagnostics.Conditional("TRACE")]
    private void TraceInfo(string format, params object[] arguments)
    {
      Trace.Log(this, format, arguments);
    }
  }
}
