// 
// Copyright (c) 2010 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Text;

using Trace = Smdn.Net.Pop3.Client.Trace;

namespace Smdn.Net.Pop3.Protocol.Transmission.Transaction {
  internal abstract class PopTransactionBase<TTransactionResult> : PopTransactionBase {
    public TTransactionResult TransactionResult { get; private set; } = default;

    protected PopTransactionBase(PopConnection connection)
      : base(connection)
    {
    }

    protected void Finish(PopStatusResponse statusResponse, TTransactionResult transactionResult)
    {
      TransactionResult = transactionResult;

      Finish(statusResponse);
    }
  }

  internal abstract class PopTransactionBase : IPopTransaction, IDisposable
  {
    public PopConnection Connection => connection;
    public IDictionary<string, string> RequestArguments => requestArguments;
    public PopCommandResult Result => result;

    private bool IsFinished => (result != null) || (connection == null /* in case of being disposed or exception */ );

    protected bool IsResponseMultiline { get; set; }
    protected PopStatusResponse StatusResponse => statusResponse;

    private const int initialRequestArgumentCapacity = 2;
    private const int initialResponseCapacity = 2;

    protected PopTransactionBase(PopConnection connection)
    {
      if (connection == null)
        throw new ArgumentNullException(nameof(connection));

      this.connection = connection;
      this.requestArguments = new Dictionary<string, string>(initialRequestArgumentCapacity, StringComparer.Ordinal);
      this.result = null;
      this.receivedResponses = new List<PopResponse>(capacity: initialResponseCapacity);
      this.IsResponseMultiline = false;
    }

    public virtual void Dispose()
    {
      // nothing to do in this class
      connection = null;
    }

    private PopConnection connection;
    private Dictionary<string, string> requestArguments;
    private PopStatusResponse statusResponse;
    private PopCommandResult result;

#region "processing transaction / state transition"
    protected abstract PopCommand PrepareCommand();

    void IPopTransaction.Process()
    {
#if DEBUG
      if (connection == null)
        throw new ObjectDisposedException(GetType().FullName);

      RejectFinished();
#endif

      connection.HandleResponseAsMultiline = false;

      // prepare and send command
      try {
        var command = PrepareCommand();

        if (command != null)
          connection.SendCommand(command);
      }
      catch (Exception ex) {
        connection = null;
        Trace.Log(ex);
        throw;
      }

      // receive response or send continuation
      try {
        for (;;) {
          ReceiveResponse();

          if (IsFinished)
            return;
        }
      }
      catch (Exception ex) {
        connection = null;
        Trace.Log(ex);
        throw;
      }
    }

    private void ReceiveResponse()
    {
      PopResponse response = null;

      try {
        for (;;) {
          response = connection.TryReceiveResponse();

          if (response != null) {
            receivedResponses.Add(response);
            break;
          }
        }
      }
      catch (PopMalformedResponseException) {
        // protocol error, abort transaction
        throw;
      }

      try {
        switch (response) {
          case PopFollowingResponse following: OnFollowingResponseReceived(following); break;
          case PopStatusResponse status: OnStatusResponseReceived(status); break;
          case PopTerminationResponse termination: OnTerminationResponse(termination); break;
          case PopContinuationRequest continuationRequest: OnContinuationRequestReceived(continuationRequest); break;
        }
      }
      catch (PopMalformedTextException ex) {
        if (response is PopStatusResponse) {
          // error in response data, abort transaction
          throw;
        }
        else {
          // error in response data, continue processing
          Trace.Log(ex);
        }
      }
    }

    protected void Finish(PopStatusResponse statusResponse)
    {
#if DEBUG
      RejectFinished();
#endif

      result = new PopCommandResult(statusResponse, receivedResponses);
    }

#if DEBUG
    protected Exception CreateRequestArgumentsException(string message)
    {
      return new InvalidOperationException(message);
    }

    private void RejectFinished()
    {
      if (IsFinished)
        throw new InvalidOperationException("transaction already finished");
    }
#endif

    private List<PopResponse> receivedResponses;
#endregion

#region "response processing"
    protected virtual void OnFollowingResponseReceived(PopFollowingResponse following)
    {
    }

    protected virtual void OnStatusResponseReceived(PopStatusResponse status)
    {
      statusResponse = status;

      if (IsResponseMultiline && status.Status == PopStatusIndicator.Positive)
        connection.HandleResponseAsMultiline = true; // receive following responses
      else
        Finish(status);
    }

    protected virtual void OnTerminationResponse(PopTerminationResponse termination)
    {
      if (IsResponseMultiline)
        Finish(statusResponse);
    }

    protected virtual void OnContinuationRequestReceived(PopContinuationRequest continuationRequest)
    {
    }
#endregion

    public override string ToString()
    {
      return StringConversion.ToString(GetType(), new (string, object)[] {
        (nameof(Connection), Connection),
        (nameof(RequestArguments), RequestArguments),
        (nameof(Result), Result),
      });
    }
  }
}
