// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;

using Smdn.Net.MessageAccessProtocols.Strings;
using Smdn.Net.Pop3.Protocol.Strings;
using Smdn.Security.Authentication.Sasl;

namespace Smdn.Net.Pop3.Protocol {
  // enum types:
  //   PopStringEnum
  //     => handles string constants
  //     PopAuthenticationMechanism
  //       => handles authentication mechanisms
  //   * PopCapability
  //       => handles capability constants

#if SERIALIZATION
  [Serializable]
#endif
  public sealed class PopCapability : PopEnumValue, IEquatable<PopCapability> {
    private static readonly PopCapabilitySet allCapabilities;
    public static IReadOnlyStringSet<PopCapability> AllCapabilities => allCapabilities;

    /*
     * POP3 Extension Mechanism
     * http://www.iana.org/assignments/pop3-extension-mechanism
     */
    public static readonly PopCapability /* [RFC2449] POP3 Extension Mechanism */
                                      Top = new PopCapability("TOP");
    public static readonly PopCapability /* [RFC2449] POP3 Extension Mechanism */
                                      User = new PopCapability("USER");
    public static readonly PopCapability /* [RFC2449] POP3 Extension Mechanism */
                                      Sasl = new PopCapability("SASL");
    public static readonly PopCapability /* [RFC2449] POP3 Extension Mechanism */
                                      RespCodes = new PopCapability("RESP-CODES");
    public static readonly PopCapability /* [RFC2449] POP3 Extension Mechanism */
                                      LoginDelay = new PopCapability("LOGIN-DELAY");
    public static readonly PopCapability /* [RFC2449] POP3 Extension Mechanism */
                                      Pipelining = new PopCapability("PIPELINING");
    public static readonly PopCapability /* [RFC2449] POP3 Extension Mechanism */
                                      Expire = new PopCapability("EXPIRE");
    public static readonly PopCapability /* [RFC2449] POP3 Extension Mechanism */
                                      Uidl = new PopCapability("UIDL");
    public static readonly PopCapability /* [RFC2449] POP3 Extension Mechanism */
                                      Implementation = new PopCapability("IMPLEMENTATION");
    public static readonly PopCapability /* [RFC2595] Using TLS with IMAP, POP3 and ACAP */
                                      Stls = new PopCapability("STLS");
    public static readonly PopCapability /* [RFC3206] The SYS and AUTH POP Response Codes */
                                      AuthRespCode = new PopCapability("AUTH-RESP-CODE");
    public static readonly PopCapability /* [RFC6856] Post Office Protocol Version 3 (POP3) Support for UTF-8 */
                                      Utf8 = new PopCapability("UTF8");
    public static readonly PopCapability /* [RFC6856] Post Office Protocol Version 3 (POP3) Support for UTF-8 */
                                      Lang = new PopCapability("LANG");

    /*
     * draft
     */

    /*
     * extended capabilities
     */

    static PopCapability()
    {
      var capabilities = new List<PopCapability>(CreateDefinedConstantsSet<PopCapability>());

      foreach (var saslMechansim in SaslMechanisms.AllMechanisms) {
        capabilities.Add(new PopCapability("SASL", saslMechansim));
      }

      allCapabilities = new PopCapabilitySet(capabilities, asReadOnly: true);
    }

    internal static PopCapability GetKnownOrCreate(string capability, string[] arguments)
    {
      if (allCapabilities.TryGetValue(capability, out var knownCapability)) {
        if (knownCapability.ContainsAllArguments(arguments))
          return knownCapability;
      }

      //Trace.Verbose("unknown capability: {0} {1}", capability, string.Join(" ", arguments));
      return new PopCapability(capability, arguments);
    }

    /*
     * instance members
     */
    public string Tag { get; }
    public IReadOnlyList<string> Arguments { get; }
    public override string Value {
      get
      {
        if (Arguments.Count == 0)
          return Tag;
        else
          return string.Concat(Tag, " ", string.Join(" ", Arguments));
      }
    }

    public PopCapability(string tag, params string[] arguments)
      : base(tag)
    {
      if (tag == null)
        throw new ArgumentNullException(nameof(tag));
      if (tag.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString(nameof(tag));

      arguments = arguments
#if NET45
        ?? ArrayExtensions.Empty<string>();
#else
        ?? Array.Empty<string>();
#endif

      foreach (var arg in arguments) {
        if (arg == null)
          throw ExceptionUtils.CreateAllItemsOfArgumentMustBeNonNull(nameof(arguments));
        if (arg.Length == 0)
          throw new ArgumentException("argument must be non-empty", nameof(arguments));
      }

      Tag = tag;
      Arguments = arguments;
    }

    public bool ContainsAllArguments(params string[] arguments)
      => ContainsAllArguments((IEnumerable<string>)arguments);

    public bool ContainsAllArguments(IEnumerable<string> arguments)
    {
      if (arguments == null)
        throw new ArgumentNullException(nameof(arguments));

      return !arguments.Except(Arguments, StringComparer.OrdinalIgnoreCase).Any();
    }

    #region "equatable"
    public static bool operator ==(PopCapability x, PopCapability y) => Equals(x, y);
    public static bool operator !=(PopCapability x, PopCapability y) => !Equals(x, y);
    public bool Equals(PopCapability other) => Equals(this, other);
    public override bool Equals(object obj) => obj is PopCapability other ? Equals(this, other) : base.Equals(obj);
    public override int GetHashCode() => base.GetHashCode();
    #endregion
  }
}
