using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Threading.Tasks;
using NUnit.Framework;

using Smdn.Net.Connection;
using Smdn.Net.Pop3.Protocol.Transmission;

namespace Smdn.Net.Pop3.Protocol.Session {
  [TestFixture]
  public class PopSessionTests {
    private void AssertSessionNotConnected(PopSession session)
    {
      Assert.AreEqual(PopSessionState.NotConnected, session.State);
      Assert.IsFalse(session.IsTransactionProceeding);

      Assert.Throws<InvalidOperationException>(() => { Assert.Fail("IsSecureConnection = {0}", session.IsSecureConnection); });

      Assert.Throws<InvalidOperationException>(() => session.Capa());
      Assert.AreEqual(PopSessionState.NotConnected, session.State);

      Assert.Throws<InvalidOperationException>(() => session.User("user"));
      Assert.AreEqual(PopSessionState.NotConnected, session.State);

      Assert.Throws<InvalidOperationException>(() => session.NoOp());
      Assert.AreEqual(PopSessionState.NotConnected, session.State);

      Assert.DoesNotThrow(() => { session.ReceiveTimeout = 1; });
      Assert.DoesNotThrow(() => { Assert.AreEqual(1, session.ReceiveTimeout); });
      Assert.DoesNotThrow(() => { session.SendTimeout = 1; });
      Assert.DoesNotThrow(() => { Assert.AreEqual(1, session.SendTimeout); });
    }

    [Test]
    public void TestConstruct()
    {
      using (var session = new PopSession()) {
        AssertSessionNotConnected(session);
      }
    }

    internal static void AssertSessionDisposed(PopSession session, PopPseudoServer server, bool disposedByTransactionTimeout = false)
    {
      Assert.AreEqual(PopSessionState.Disconnected, session.State);

      if (!disposedByTransactionTimeout)
        Assert.IsFalse(session.IsTransactionProceeding);

      // attempt to send command
      Assert.Throws<ObjectDisposedException>(() => session.NoOp());

      // attempt to reconnect
      if (server != null) {
        Assert.Throws<ObjectDisposedException>(() => session.ConnectAsync(server.Host, server.Port));
        Assert.AreEqual(PopSessionState.Disconnected, session.State);

        if (!disposedByTransactionTimeout)
          Assert.IsFalse(session.IsTransactionProceeding);
      }

      // attempt to re-disconnect
      Assert.Throws<ObjectDisposedException>(() => session.Disconnect());

      // attempt to re-dispose
      Assert.DoesNotThrow(() => session.Dispose());

      Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("SendTimeout = {0}", session.SendTimeout); });
      Assert.Throws<ObjectDisposedException>(() => { session.SendTimeout = 0; });
      Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("ReceiveTimeout = {0}", session.ReceiveTimeout); });
      Assert.Throws<ObjectDisposedException>(() => { session.ReceiveTimeout = 0; });
      Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("TransactionTimeout = {0}", session.TransactionTimeout); });
      Assert.Throws<ObjectDisposedException>(() => { session.TransactionTimeout = 0; });
      Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("IsSecureConnection = {0}", session.IsSecureConnection); });
      Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("Authority = {0}", session.Authority); });
    }

    [Test]
    public void TestNotConnected_Dispose()
    {
      using (var session = new PopSession()) {
        Assert.DoesNotThrow(() => session.Dispose());

        AssertSessionDisposed(session, null);
      }
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestNotConnected_Disconnect(bool logout)
    {
      using (var session = new PopSession()) {
        Assert.DoesNotThrow(() => session.Disconnect(logout));

        AssertSessionDisposed(session, null);
      }
    }

    [Test]
    public void TestConnectAsync()
    {
      PopPseudoServer.Start((server) => {
        using (var session = new PopSession(server.Host, server.Port)) {
          Assert.IsFalse(session.ApopAvailable);
          Assert.IsTrue(string.IsNullOrEmpty(session.Timestamp));
          Assert.AreEqual(PopSessionState.Authorization, session.State);
          Assert.AreEqual(string.Format("pop://{0}:{1}/", server.Host, server.Port),
                          session.Authority.AbsoluteUri);
          Assert.AreEqual(0, session.ServerCapabilities.Count);
        }
      });
    }

    [Test]
    public void TestConnectAsync_ConnectionFailure_Cancellation_Connection()
    {
      var timeout = TimeSpan.FromMilliseconds(250);
      var timeoutAcceptable = TimeSpan.FromMilliseconds(500);

      using (var session = new PopSession()) {
        TestUtils.Assert.Elapses(timeout, timeoutAcceptable, () => {
          Assert.ThrowsAsync<OperationCanceledException>(async () => {
            using (var cts = new CancellationTokenSource(timeout)) {
              await session.ConnectAsync("pop.nonexistent.test", 110, cancellationToken: cts.Token);
            }
          });
        });

        AssertSessionNotConnected(session);
      }
    }

    [Test]
    public void TestConnectAsync_ConnectionFailure_SocketException()
    {
      using (var session = new PopSession()) {
        var ex = Assert.ThrowsAsync<PopConnectionException>(async () => await session.ConnectAsync("pop.invalid", 10110));

        Assert.IsInstanceOf<SocketException>(ex.InnerException);

        AssertSessionNotConnected(session);
      }
    }

    [Test]
    public void TestConnectAsync_TransactionFailure_Timeout_ServerNotAccepting()
    {
      var timeout = TimeSpan.FromMilliseconds(250);
      var timeoutAcceptable = TimeSpan.FromMilliseconds(500);

      PopPseudoServer.StartNotAccepting((server) => {
        using (var session = new PopSession()) {
          session.TransactionTimeout = (int)timeout.TotalMilliseconds;

          TestUtils.Assert.Elapses(timeout, timeoutAcceptable, () => {
            var ex = Assert.ThrowsAsync<PopConnectionException>(async () => await session.ConnectAsync(server.Host, server.Port));

            Assert.IsInstanceOf<TimeoutException>(ex.InnerException);
          });

          AssertSessionDisposed(session, server, disposedByTransactionTimeout: true);
        }
      });
    }

    [Test]
    public void TestConnectAsync_TransactionFailure_Timeout_NoGreetingResponse()
    {
      var timeout = TimeSpan.FromMilliseconds(250);
      var timeoutAcceptable = TimeSpan.FromMilliseconds(500);

      PopPseudoServer.Start((server) => {
        server.SetGreetingResponse(null); // no greeting response

        using (var session = new PopSession()) {
          session.TransactionTimeout = (int)timeout.TotalMilliseconds;

          TestUtils.Assert.Elapses(timeout, timeoutAcceptable, () => {
            var ex = Assert.ThrowsAsync<PopConnectionException>(async () => await session.ConnectAsync(server.Host, server.Port));

            Assert.IsInstanceOf<TimeoutException>(ex.InnerException);
          });

          AssertSessionDisposed(session, server, disposedByTransactionTimeout: true);
        }
      });
    }

    [Test]
    public void TestConnectAsync_TransactionFailure_ByeResponse()
    {
      PopPseudoServer.Start((server) => {
        server.SetGreetingResponse("-ERR try again\r\n", null, null, true);

        using (var session = new PopSession()) {
          var ex = Assert.ThrowsAsync<PopConnectionException>(async () => await session.ConnectAsync(server.Host, server.Port));

          Assert.IsNull(ex.InnerException);
          StringAssert.Contains("try again", ex.Message);

          AssertSessionDisposed(session, server);
        }
      });
    }

    [TestCase(".ready\r\n")] // following response
    [TestCase(".\r\n")] // termination response
    [TestCase("+\r\n")] // continuation request
    [TestCase("+ ready\r\n")] // continuation request
    [TestCase("* OK pseudo IMAP server ready\r\n")] // treated as following response
    public void TestConnectAsync_TransactionFailure_UnexpectedResponse(string response)
    {
      PopPseudoServer.Start((server) => {
        server.SetGreetingResponse(response);

        using (var session = new PopSession()) {
          var ex = Assert.ThrowsAsync<PopConnectionException>(async () => await session.ConnectAsync(server.Host, server.Port));

          Assert.IsNull(ex.InnerException);
          StringAssert.Contains("session establishment failed (unexpected", ex.Message);

          AssertSessionDisposed(session, server);
        }
      });
    }

    [TestCase("+OK POP3 server ready <msg-id\r\n")]
    [TestCase("+OK POP3 server ready >msg-id<\r\n")]
    [TestCase("+OK []<\r\n")]
    public void TestConnectAsync_TransactionFailure_MalformedResponse(string response)
    {
      PopPseudoServer.Start((server) => {
        server.SetGreetingResponse(response);

        using (var session = new PopSession()) {
          var ex = Assert.ThrowsAsync<PopConnectionException>(async () => await session.ConnectAsync(server.Host, server.Port));

          Assert.IsNotNull(ex.InnerException);
          Assert.IsInstanceOf<PopMalformedResponseException>(ex.InnerException);
          StringAssert.Contains("session establishment failed (malformed or invalid response)", ex.Message);

          AssertSessionDisposed(session, server);
        }
      });
    }

    [Test]
    public void TestConnect_Ssl()
    {
      PopPseudoServer.Start((server) => {
        var streamUpgraded = false;

        using (var session = new PopSession(server.Host,
                                            server.Port,
                                            createAuthenticatedStream: (connection, baseStream, cancellationToken) => { streamUpgraded = true; return Task.FromResult<Stream>(baseStream); /* TODO: return SslStream */ })) {
          Assert.IsFalse(session.ApopAvailable);
          Assert.IsTrue(string.IsNullOrEmpty(session.Timestamp));
          Assert.AreEqual(PopSessionState.Authorization, session.State);
          Assert.AreEqual(string.Format("pops://{0}:{1}/", server.Host, server.Port),
                          session.Authority.AbsoluteUri);
          Assert.AreEqual(0, session.ServerCapabilities.Count);
          Assert.IsTrue(streamUpgraded, "stream upgraded");
          Assert.IsTrue(session.IsSecureConnection);
        }
      });
    }

    [Test]
    public void TestConnectAsync_Ssl_ExceptionWhileUpgrading()
    {
      PopPseudoServer.Start((server) => {
        using (var session = new PopSession()) {
          var ex = Assert.ThrowsAsync<PopSecureConnectionException>(async () => {
            await session.ConnectAsync(server.Host,
                                       server.Port,
                                       createAuthenticatedStream: (connection, baseStream, cancellationToken) => throw new System.Security.Authentication.AuthenticationException());
          });

          Assert.IsNotNull(ex.InnerException);
          Assert.IsInstanceOf<System.Security.Authentication.AuthenticationException>(ex.InnerException);

          AssertSessionNotConnected(session);
        }
      });
    }

    [Test]
    public void TestConnectAsync_ResolveHostAddressCallback()
    {
      PopPseudoServer.Start((server) => {
        using (var session = new PopSession()) {
          const string pseudoServerHostName  = "pop.invalid";

          var resolveHostAddress = new ResolveHostAddressCallback((hostname, cancellationToken) => {
            Assert.AreEqual(pseudoServerHostName, hostname);

            return Task.FromResult(server.ServerEndPoint.Address);
          });

          server.SetGreetingResponse("+OK PopPseudoServer ready\r\n");

          Assert.DoesNotThrowAsync(async () => {
            await session.ConnectAsync(pseudoServerHostName,
                                       server.Port,
                                       resolveHostAddress: resolveHostAddress);
          });

          Assert.AreEqual(PopSessionState.Authorization, session.State);
          Assert.AreEqual($"pop://{pseudoServerHostName}:{server.Port}/", session.Authority.AbsoluteUri);
          Assert.IsFalse(session.IsSecureConnection);
        }
      });
    }

    [Test]
    public void TestConnectAsync_ResolveHostAddressCallback_Exception()
    {
      using (var session = new PopSession()) {
        const string pseudoServerHostName  = "pop.invalid";

        var resolveHostAddress = new ResolveHostAddressCallback((hostname, cancellationToken) => {
          Assert.AreEqual(pseudoServerHostName, hostname);
#if NET45
          var tcs = new TaskCompletionSource<IPAddress>();
          tcs.SetException(new NotImplementedException());
          return tcs.Task;
#else
          return Task.FromException<IPAddress>(new NotImplementedException());
#endif
        });

        var ex = Assert.ThrowsAsync<PopConnectionException>(async () => {
          await session.ConnectAsync(pseudoServerHostName,
                                     resolveHostAddress: resolveHostAddress);
        });

        Assert.IsInstanceOf<NotImplementedException>(ex.InnerException);

        AssertSessionNotConnected(session);
      }
    }

    [Test]
    public void TestDispose() => DisconnectDispose(session => session.Dispose());

    [Test]
    public void TestDisconnect() => DisconnectDispose(session => session.Disconnect(logout: false));

    private void DisconnectDispose(Action<PopSession> disposeSession)
    {
      TestCases.WithConnectedSession((session, server) => {
        disposeSession(session);

        AssertSessionDisposed(session, server);
      });
    }

    [Test]
    public void TestTransactionException_TransactionTimeout()
    {
      TestCases.WithConnectedSession((session, server) => {
        session.TransactionTimeout = 250;
        session.SendTimeout = System.Threading.Timeout.Infinite;
        session.ReceiveTimeout = System.Threading.Timeout.Infinite;

        // not respond CAPA response termination
        server.EnqueueResponse("CAPA\r\n",
                               "+OK List of capabilities follows\r\n");

        Assert.Throws<TimeoutException>(() => session.Capa());

        AssertSessionDisposed(session, server, disposedByTransactionTimeout: true);
      });
    }

    [Test]
    public void TestTransactionException_SocketTimeout()
    {
      TestCases.WithConnectedSession((session, server) => {
        session.TransactionTimeout = System.Threading.Timeout.Infinite;
        session.SendTimeout = 250;
        session.ReceiveTimeout = 250;

        // not respond CAPA
        //server.EnqueueResponse("+OK List of capabilities follows\r\n.\r\n");

        Assert.Throws<TimeoutException>(() => session.Capa());

        AssertSessionDisposed(session, server);
      });
    }

    [Test]
    public void TestTransactionException_InternalError()
    {
      TestCases.WithLoggedInSession((session, server) => {
        server.EnqueueResponse("LIST 1\r\n",
                               "+OK 1 TEXT\r\n" /* invalid response */);

        var ex = Assert.Throws<PopMalformedTextException>(() => session.List(1, out _));

        AssertSessionDisposed(session, server);
      });
    }

    [Test]
    public void TestDisconnectedFromServer()
    {
      PopPseudoServer.Start((server) => {
        server.SetGreetingResponse("+OKr\n", null, null, true);

        using (var session = new PopSession(server.Host, server.Port)) {
          Assert.Throws<PopConnectionException>(() => session.Quit());

          AssertSessionDisposed(session, server);
        }
      });
    }

    [Test]
    public void TestGreetingWithTimestamp()
    {
      PopPseudoServer.Start((server) => {
        server.SetGreetingResponse("+OK POP3 server ready <1896.697170952@dbc.mtview.ca.us>\r\n");

        using (var session = new PopSession(server.Host, server.Port)) {
          Assert.IsTrue(session.ApopAvailable);
          Assert.AreEqual("<1896.697170952@dbc.mtview.ca.us>", session.Timestamp);
          Assert.AreEqual(PopSessionState.Authorization, session.State);
          Assert.AreEqual(string.Format("pop://{0}:{1}/", server.Host, server.Port),
                          session.Authority.AbsoluteUri);
          Assert.AreEqual(0, session.ServerCapabilities.Count);
        }
      });
    }

    [Test]
    public void TestGreetingWithTimestamp_RespondOnlyTimestamp()
    {
      PopPseudoServer.Start((server) => {
        server.SetGreetingResponse("+OK <1896.697170952@dbc.mtview.ca.us>\r\n");

        using (var session = new PopSession(server.Host, server.Port)) {
          Assert.IsTrue(session.ApopAvailable);
          Assert.AreEqual("<1896.697170952@dbc.mtview.ca.us>", session.Timestamp);
          Assert.AreEqual(PopSessionState.Authorization, session.State);
          Assert.AreEqual(string.Format("pop://{0}:{1}/", server.Host, server.Port),
                          session.Authority.AbsoluteUri);
          Assert.AreEqual(0, session.ServerCapabilities.Count);
        }
      });
    }

    [Test]
    public void TestSetReceiveTimeoutZero()
    {
      TestCases.WithConnectedSession(session => {
        session.ReceiveTimeout = 0;

        Assert.AreEqual(0, session.ReceiveTimeout);
      });
    }

    [Test]
    public void TestSetReceiveTimeoutInfinite()
    {
      TestCases.WithConnectedSession(session => {
        session.ReceiveTimeout = Timeout.Infinite;

        Assert.AreEqual(Timeout.Infinite, session.ReceiveTimeout);
      });
    }

    [Test]
    public void TestSetSendTimeoutZero()
    {
      TestCases.WithConnectedSession(session => {
        session.SendTimeout = 0;

        Assert.AreEqual(0, session.SendTimeout);
      });
    }

    [Test]
    public void TestSetSendTimeoutInfinite()
    {
      TestCases.WithConnectedSession(session => {
        session.SendTimeout = Timeout.Infinite;

        Assert.AreEqual(Timeout.Infinite, session.SendTimeout);
      });
    }

    [Test]
    public void TestSetTransactionTimeoutZero()
    {
      TestCases.WithConnectedSession(session => {
        session.TransactionTimeout = 0;

        Assert.AreEqual(0, session.TransactionTimeout);
      });
    }

    [Test]
    public void TestSetTransactionTimeoutInfinite()
    {
      TestCases.WithConnectedSession(session => {
        session.TransactionTimeout = Timeout.Infinite;

        Assert.AreEqual(Timeout.Infinite, session.TransactionTimeout);
      });
    }

    [Test]
    public void TestBeginNewTransactionWhileTransactionProceeding()
    {
      TestCases.WithConnectedSession((session, server) => {
        using (var transactionStartedEvent = new ManualResetEvent(false)) {
          server.EnqueueResponse("CAPA\r\n",
                                 "+OK Capability list follows\r\n", // incomplete response
                                 null,
                                 () => transactionStartedEvent.Set());

          var capaTask = Task.Run(() => session.Capa());

          transactionStartedEvent.WaitOne();

          Assert.IsTrue(session.IsTransactionProceeding);

          Assert.Throws<InvalidOperationException>(() => session.Capa());

          server.EnqueuePushResponse(".\r\n");

          Assert.DoesNotThrow(() => {
            if (!capaTask.Wait(TimeSpan.FromSeconds(1)))
              Assert.Fail("transaction not stopped");
          });
        }
      });
    }

    [Test]
    public void TestGenericCommandWithNoArguments()
    {
      TestCases.WithConnectedSession((session, server) => {
        // generic command transaction
        server.EnqueueResponse("X-EXT-COMMAND\r\n",
                               "+OK done\r\n");

        Assert.IsTrue(session.GenericCommand("X-EXT-COMMAND", out var responses).IsSuccess);

        Assert.AreEqual(1, responses.Count);
        Assert.IsInstanceOf<PopStatusResponse>(responses[0]);

        var status = responses[0] as PopStatusResponse;

        Assert.AreEqual("done", status.Text);
      });
    }

    [Test]
    public void TestGenericCommandWithArguments()
    {
      TestCases.WithConnectedSession((session, server) => {
        // generic command transaction
        server.EnqueueResponse("X-EXT-COMMAND arg1 arg2 arg3\r\n",
                               "+OK done\r\n");

        Assert.IsTrue(session.GenericCommand("X-EXT-COMMAND", out var responses, "arg1", "arg2", "arg3").IsSuccess);

        Assert.AreEqual(1, responses.Count);
        Assert.IsInstanceOf<PopStatusResponse>(responses[0]);

        var status = responses[0] as PopStatusResponse;

        Assert.AreEqual("done", status.Text);
      });
    }

    [Test]
    public void TestGenericCommandMultilineResponse()
    {
      TestCases.WithConnectedSession((session, server) => {
        // generic command transaction
        server.EnqueueResponse("X-EXT-COMMAND\r\n",
                               "+OK [X-RESP-CODE] done\r\n" +
                               "1st line\r\n" +
                               "2nd line\r\n" +
                               ".byte-stuffed line\r\n" +
                               ".\r\n");

        Assert.IsTrue(session.GenericCommand("X-EXT-COMMAND", true, out var responses).IsSuccess);

        Assert.AreEqual(5, responses.Count);
        Assert.IsInstanceOf<PopStatusResponse>(responses[0]);
        Assert.IsInstanceOf<PopFollowingResponse>(responses[1]);
        Assert.IsInstanceOf<PopFollowingResponse>(responses[2]);
        Assert.IsInstanceOf<PopFollowingResponse>(responses[3]);
        Assert.IsInstanceOf<PopTerminationResponse>(responses[4]);

        var status = responses[0] as PopStatusResponse;

        Assert.AreEqual("done", status.Text);
        Assert.IsNotNull(status.ResponseText);
        Assert.AreEqual("X-RESP-CODE", status.ResponseText.Code.Value);

        Assert.AreEqual("1st line", (responses[1] as PopFollowingResponse).Text.ToString());
        Assert.AreEqual("2nd line", (responses[2] as PopFollowingResponse).Text.ToString());
        Assert.AreEqual("byte-stuffed line", (responses[3] as PopFollowingResponse).Text.ToString());
      });
    }
  }
}
