using System;
using NUnit.Framework;

namespace Smdn.Net.Pop3.Protocol.Session {
  [TestFixture]
  public class PopSessionCommandsAnyStateTests {
    [Test]
    public void TestCapa()
    {
      TestCases.WithConnectedSession((session, server) => {
        Assert.IsNotNull(session.ServerCapabilities);
        Assert.AreEqual(0, session.ServerCapabilities.Count);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK Capability list follows\r\n" +
                               "TOP\r\n" +
                               "USER\r\n" +
                               "SASL CRAM-MD5 KERBEROS_V4\r\n" +
                               "RESP-CODES\r\n" +
                               "LOGIN-DELAY 900\r\n" +
                               "PIPELINING\r\n" +
                               "EXPIRE 60\r\n" +
                               "UIDL\r\n" +
                               "IMPLEMENTATION Shlemazle-Plotz-v302\r\n" +
                               ".\r\n");

        Assert.IsTrue(session.Capa(out var capabilities).IsSuccess);

        Assert.AreEqual(9, capabilities.Count);
        Assert.IsTrue(capabilities.Contains(PopCapability.Top));
        Assert.IsTrue(capabilities.Contains(PopCapability.User));
        Assert.IsTrue(capabilities.Contains(new PopCapability("SASL", "CRAM-MD5", "KERBEROS_V4")));
        Assert.IsTrue(capabilities.Contains(PopCapability.RespCodes));
        Assert.IsTrue(capabilities.Contains(new PopCapability(PopCapability.LoginDelay.Value, "900")));
        Assert.IsTrue(capabilities.Contains(PopCapability.Pipelining));
        Assert.IsTrue(capabilities.Contains(new PopCapability(PopCapability.Expire.Value, "60")));
        Assert.IsTrue(capabilities.Contains(PopCapability.Uidl));
        Assert.IsTrue(capabilities.Contains(new PopCapability(PopCapability.Implementation.Value, "Shlemazle-Plotz-v302")));

        Assert.IsTrue(capabilities.IsCapable(PopCapability.Sasl));
        Assert.IsTrue(capabilities.IsCapable(PopCapability.LoginDelay));
        Assert.IsTrue(capabilities.IsCapable(PopCapability.Expire));
        Assert.IsTrue(capabilities.IsCapable(PopCapability.Implementation));
        Assert.IsTrue(capabilities.IsCapable(PopAuthenticationMechanism.CRAMMD5));
        Assert.IsTrue(capabilities.IsCapable(PopAuthenticationMechanism.KerberosV4));

        Assert.IsNotNull(session.ServerCapabilities);
        Assert.AreEqual(9, session.ServerCapabilities.Count);
        Assert.IsTrue(session.ServerCapabilities.Contains(PopCapability.User));
        Assert.IsTrue(session.ServerCapabilities.IsCapable(PopCapability.Sasl));
      });
    }

    [Test]
    public void TestQuitInAuthorizationState()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueResponse("QUIT\r\n",
                               "+OK dewey POP3 server signing off\r\n",
                               isQuitResponse: true);

        Assert.IsTrue(session.Quit().IsSuccess);

        PopSessionTests.AssertSessionDisposed(session, server);
      });
    }

    [Test]
    public void TestQuitInTransactionState()
    {
      TestCases.WithLoggedInSession((session, server) => {
        server.EnqueueResponse("QUIT\r\n",
                               "+OK dewey POP3 server signing off (maildrop empty)\r\n",
                               isQuitResponse: true);

        Assert.IsTrue(session.Quit().IsSuccess);

        PopSessionTests.AssertSessionDisposed(session, server);
      });
    }

    [Test]
    public void TestQuitAfterDisconnect() => QuitAfterDisconnectOrDispose(session => session.Disconnect(logout: false));

    [Test]
    public void TestQuitAfterDispose() => QuitAfterDisconnectOrDispose(session => session.Dispose());

    private void QuitAfterDisconnectOrDispose(Action<PopSession> disposeAction)
    {
      TestCases.WithConnectedSession((session, server) => {
        disposeAction(session);

        Assert.Throws<ObjectDisposedException>(() => session.Quit());

        PopSessionTests.AssertSessionDisposed(session, server);
      });
    }

    [Test]
    public void TestQuitNotConnected()
    {
      using (var session = new PopSession()) {
        Assert.Throws<InvalidOperationException>(() => session.Quit());
      }
    }
  }
}
